from pathlib import Path
from typing import Optional, Union

import k4a_module

from .capture import PyK4ACapture
from .config import Config
from .errors import K4AException
from .pyk4a import PyK4A
from .results import Result


class PyK4ARecord:
    def __init__(
        self, path: Union[str, Path], config: Config, device: Optional[PyK4A] = None, thread_safe: bool = True
    ):
        self._path: Path = Path(path)
        self.thread_safe = thread_safe
        self._device: Optional[PyK4A] = device
        self._config: Config = config
        self._handle: Optional[object] = None
        self._header_written: bool = False
        self._captures_count: int = 0

    def __del__(self):
        if self.created:
            self.close()

    def create(self) -> None:
        """ Create record file """
        if self.created:
            raise K4AException(f"Record already created {self._path}")
        device_handle = self._device._device_handle if self._device else None
        result, handle = k4a_module.record_create(
            device_handle, str(self._path), self.thread_safe, *self._config.unpack()
        )
        if result != Result.Success:
            raise K4AException(f"Cannot create record {self._path}")
        self._handle = handle

    def close(self):
        """ Close record """
        self._validate_is_created()
        k4a_module.record_close(self._handle, self.thread_safe)
        self._handle = None

    def write_header(self):
        """ Write MKV header """
        self._validate_is_created()
        if self.header_written:
            raise K4AException(f"Header already written {self._path}")
        result: Result = k4a_module.record_write_header(self._handle, self.thread_safe)
        if result != Result.Success:
            raise K4AException(f"Cannot write record header {self._path}")
        self._header_written = True

    def write_capture(self, capture: PyK4ACapture):
        """ Write capture to file (send to queue) """
        self._validate_is_created()
        if not self.header_written:
            self.write_header()
        result: Result = k4a_module.record_write_capture(self._handle, capture._capture_handle, self.thread_safe)
        if result != Result.Success:
            raise K4AException(f"Cannot write capture {self._path}")
        self._captures_count += 1

    def flush(self):
        """ Flush queue"""
        self._validate_is_created()
        result: Result = k4a_module.record_flush(self._handle, self.thread_safe)
        if result != Result.Success:
            raise K4AException(f"Cannot flush data {self._path}")

    @property
    def created(self) -> bool:
        return self._handle is not None

    @property
    def header_written(self) -> bool:
        return self._header_written

    @property
    def captures_count(self) -> int:
        return self._captures_count

    @property
    def path(self) -> Path:
        return self._path

    def _validate_is_created(self):
        if not self.created:
            raise K4AException("Record not created.")



from .results import Result


class K4AException(Exception):
    pass


class K4ATimeoutException(K4AException):
    pass


def _verify_error(res: int):
    """
    Validate k4a_module result
    """
    res = Result(res)
    if res == Result.Failed:
        raise K4AException()
    elif res == Result.Timeout:
        raise K4ATimeoutException()



from enum import IntEnum


# k4a_wait_result_t
class Result(IntEnum):
    Success = 0
    Failed = 1
    Timeout = 2


# k4a_buffer_result_t
class BufferResult(IntEnum):
    Success = 0
    Failed = 1
    TooSmall = 2


# k4a_stream_result_t
class StreamResult(IntEnum):
    Success = 0
    Failed = 1
    EOF = 2



from enum import IntEnum
from typing import Optional, Tuple

import numpy as np

import k4a_module

from .config import ColorResolution, DepthMode
from .errors import K4AException, _verify_error


class CalibrationType(IntEnum):
    UNKNOWN = -1  # Unknown
    DEPTH = 0  # Depth Camera
    COLOR = 1  # Color Sensor
    GYRO = 2  # Gyroscope
    ACCEL = 3  # Accelerometer
    NUM = 4  # Number of types excluding unknown type


class Calibration:
    def __init__(
        self, handle: object, depth_mode: DepthMode, color_resolution: ColorResolution, thread_safe: bool = True
    ):
        self._calibration_handle = handle
        self._transformation_handle: Optional[object] = None
        self.thread_safe = thread_safe
        self._depth_mode = depth_mode
        self._color_resolution = color_resolution
        self._raw: Optional[str] = None

    @classmethod
    def from_raw(
        cls, value: str, depth_mode: DepthMode, color_resolution: ColorResolution, thread_safe: bool = True
    ) -> "Calibration":
        res, handle = k4a_module.calibration_get_from_raw(thread_safe, value, depth_mode, color_resolution)
        _verify_error(res)
        return Calibration(
            handle=handle, depth_mode=depth_mode, color_resolution=color_resolution, thread_safe=thread_safe
        )

    @property
    def depth_mode(self) -> DepthMode:
        return self._depth_mode

    @property
    def color_resolution(self) -> ColorResolution:
        return self._color_resolution

    def _convert_3d_to_3d(
        self,
        source_point_3d: Tuple[float, float, float],
        source_camera: CalibrationType,
        target_camera: CalibrationType,
    ) -> Tuple[float, float, float]:
        """
            Transform a 3d point of a source coordinate system into a 3d
            point of the target coordinate system.
            :param source_point_3d  The 3D coordinates in millimeters representing a point in source_camera.
            :param source_camera    The current camera.
            :param target_camera    The target camera.
            :return                 The 3D coordinates in millimeters representing a point in target camera.
        """
        res, target_point_3d = k4a_module.calibration_3d_to_3d(
            self._calibration_handle, self.thread_safe, source_point_3d, source_camera, target_camera,
        )

        _verify_error(res)
        return target_point_3d

    def depth_to_color_3d(self, point_3d: Tuple[float, float, float]) -> Tuple[float, float, float]:
        return self._convert_3d_to_3d(point_3d, CalibrationType.DEPTH, CalibrationType.COLOR)

    def color_to_depth_3d(self, point_3d: Tuple[float, float, float]) -> Tuple[float, float, float]:
        return self._convert_3d_to_3d(point_3d, CalibrationType.COLOR, CalibrationType.DEPTH)

    def _convert_2d_to_3d(
        self,
        source_pixel_2d: Tuple[float, float],
        source_depth: float,
        source_camera: CalibrationType,
        target_camera: CalibrationType,
    ) -> Tuple[float, float, float]:
        """
            Transform a 3d point of a source coordinate system into a 3d
            point of the target coordinate system.
            :param source_pixel_2d    The 2D coordinates in px of source_camera color_image.
            :param source_depth       Depth in mm
            :param source_camera      The current camera.
            :param target_camera      The target camera.
            :return                   The 3D coordinates in mm representing a point in target camera.
        """
        res, valid, target_point_3d = k4a_module.calibration_2d_to_3d(
            self._calibration_handle, self.thread_safe, source_pixel_2d, source_depth, source_camera, target_camera,
        )

        _verify_error(res)
        if valid == 0:
            raise ValueError(f"Coordinates {source_pixel_2d} are not valid in the calibration model")

        return target_point_3d

    def convert_2d_to_3d(
        self,
        coordinates: Tuple[float, float],
        depth: float,
        source_camera: CalibrationType,
        target_camera: Optional[CalibrationType] = None,
    ):
        """
            Transform a 2d pixel to a 3d point of the target coordinate system.
        """
        if target_camera is None:
            target_camera = source_camera
        return self._convert_2d_to_3d(coordinates, depth, source_camera, target_camera)

    def _convert_3d_to_2d(
        self,
        source_point_3d: Tuple[float, float, float],
        source_camera: CalibrationType,
        target_camera: CalibrationType,
    ) -> Tuple[float, float]:
        """
            Transform a 3d point of a source coordinate system into a 3d
            point of the target coordinate system.
            :param source_point_3d    The 3D coordinates in mm of source_camera.
            :param source_camera      The current camera.
            :param target_camera      The target camera.
            :return                   The 3D coordinates in mm representing a point in target camera.
        """
        res, valid, target_px_2d = k4a_module.calibration_3d_to_2d(
            self._calibration_handle, self.thread_safe, source_point_3d, source_camera, target_camera,
        )

        _verify_error(res)
        if valid == 0:
            raise ValueError(f"Coordinates {source_point_3d} are not valid in the calibration model")

        return target_px_2d

    def convert_3d_to_2d(
        self,
        coordinates: Tuple[float, float, float],
        source_camera: CalibrationType,
        target_camera: Optional[CalibrationType] = None,
    ):
        """
            Transform a 3d point to a 2d pixel of the target coordinate system.
        """
        if target_camera is None:
            target_camera = source_camera
        return self._convert_3d_to_2d(coordinates, source_camera, target_camera)

    @property
    def transformation_handle(self) -> object:
        if not self._transformation_handle:
            handle = k4a_module.transformation_create(self._calibration_handle, self.thread_safe)
            if not handle:
                raise K4AException("Cannot create transformation handle")
            self._transformation_handle = handle
        return self._transformation_handle

    def get_camera_matrix(self, camera: CalibrationType) -> np.ndarray:
        """
        Get the camera matrix (in OpenCV compatible format) for the color or depth camera
        """
        if camera not in [CalibrationType.COLOR, CalibrationType.DEPTH]:
            raise ValueError("Camera matrix only available for color and depth cameras.")
        params = k4a_module.calibration_get_intrinsics(self._calibration_handle, self.thread_safe, camera)
        if len(params) != 14:
            raise ValueError("Unknown camera calibration type")

        cx, cy, fx, fy = params[:4]
        return np.array([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])

    def get_distortion_coefficients(self, camera: CalibrationType) -> np.ndarray:
        """
        Get the distortion coefficients (in OpenCV compatible format) for the color or depth camera
        """
        if camera not in [CalibrationType.COLOR, CalibrationType.DEPTH]:
            raise ValueError("Distortion coefficients only available for color and depth cameras.")
        params = k4a_module.calibration_get_intrinsics(self._calibration_handle, self.thread_safe, camera)
        if len(params) != 14:
            raise ValueError("Unknown camera calibration type")

        return np.array([params[4], params[5], params[13], params[12], *params[6:10]])



import k4a_module

from .calibration import Calibration, CalibrationType
from .capture import PyK4ACapture
from .config import (
    FPS,
    ColorControlCommand,
    ColorControlMode,
    ColorResolution,
    Config,
    DepthMode,
    ImageFormat,
    WiredSyncMode,
)
from .errors import K4AException, K4ATimeoutException
from .playback import PyK4APlayback, SeekOrigin
from .pyk4a import ColorControlCapabilities, PyK4A
from .record import PyK4ARecord
from .transformation import (
    color_image_to_depth_camera,
    depth_image_to_color_camera,
    depth_image_to_color_camera_custom,
    depth_image_to_point_cloud,
)


def connected_device_count() -> int:
    return k4a_module.device_get_installed_count()


__all__ = (
    "Calibration",
    "CalibrationType",
    "FPS",
    "ColorControlCommand",
    "ColorControlMode",
    "ImageFormat",
    "ColorResolution",
    "Config",
    "DepthMode",
    "WiredSyncMode",
    "K4AException",
    "K4ATimeoutException",
    "PyK4A",
    "PyK4ACapture",
    "PyK4APlayback",
    "SeekOrigin",
    "ColorControlCapabilities",
    "color_image_to_depth_camera",
    "depth_image_to_point_cloud",
    "depth_image_to_color_camera",
    "depth_image_to_color_camera_custom",
    "PyK4ARecord",
    "connected_device_count",
)



import sys
from typing import Any, Optional, Tuple

import k4a_module

from .calibration import Calibration
from .capture import PyK4ACapture
from .config import ColorControlCommand, ColorControlMode, Config
from .errors import K4AException, _verify_error


if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class PyK4A:
    TIMEOUT_WAIT_INFINITE = -1

    def __init__(self, config: Optional[Config] = None, device_id: int = 0, thread_safe: bool = True):
        self._device_id = device_id
        self._config: Config = config if (config is not None) else Config()
        self.thread_safe = thread_safe
        self._device_handle: Optional[object] = None
        self._calibration: Optional[Calibration] = None
        self.is_running = False

    def start(self):
        """
        Open device if device not opened, then start cameras and IMU
        All-in-one function
        :return:
        """
        if not self.opened:
            self.open()
        self._start_cameras()
        self._start_imu()
        self.is_running = True

    def stop(self):
        """
        Stop cameras, IMU, ... and close device
        :return:
        """
        self._stop_imu()
        self._stop_cameras()
        self._device_close()
        self.is_running = False

    def __del__(self):
        if self.is_running:
            self.stop()
        elif self.opened:
            self.close()

    @property
    def opened(self) -> bool:
        return self._device_handle is not None

    def open(self):
        """
        Open device
        You must open device before querying any information
        """
        if self.opened:
            raise K4AException("Device already opened")
        self._device_open()

    def close(self):
        self._validate_is_opened()
        self._device_close()

    def save_calibration_json(self, path: Any):
        with open(path, "w") as f:
            f.write(self.calibration_raw)

    def load_calibration_json(self, path: Any):
        with open(path, "r") as f:
            calibration = f.read()
        self.calibration_raw = calibration

    def _device_open(self):
        res, handle = k4a_module.device_open(self._device_id, self.thread_safe)
        _verify_error(res)
        self._device_handle = handle

    def _device_close(self):
        res = k4a_module.device_close(self._device_handle, self.thread_safe)
        _verify_error(res)
        self._device_handle = None

    def _start_cameras(self):
        res = k4a_module.device_start_cameras(self._device_handle, self.thread_safe, *self._config.unpack())
        _verify_error(res)

    def _start_imu(self):
        res = k4a_module.device_start_imu(self._device_handle, self.thread_safe)
        _verify_error(res)

    def _stop_cameras(self):
        res = k4a_module.device_stop_cameras(self._device_handle, self.thread_safe)
        _verify_error(res)

    def _stop_imu(self):
        res = k4a_module.device_stop_imu(self._device_handle, self.thread_safe)
        _verify_error(res)

    def get_capture(self, timeout=TIMEOUT_WAIT_INFINITE,) -> "PyK4ACapture":
        """
        Fetch a capture from the device and return a PyK4ACapture object. Images are
        lazily fetched.

        Arguments:
            :param timeout: Timeout in ms. Default is infinite.

        Returns:
            :return capture containing requested images and infos if they are available
                in the current capture. There are no guarantees that the returned
                object will contain all the requested images.

        If using any ImageFormat other than ImageFormat.COLOR_BGRA32, the color color_image must be
        decoded. See example/color_formats.py
        """
        self._validate_is_opened()
        res, capture_capsule = k4a_module.device_get_capture(self._device_handle, self.thread_safe, timeout)
        _verify_error(res)

        capture = PyK4ACapture(
            calibration=self.calibration,
            capture_handle=capture_capsule,
            color_format=self._config.color_format,
            thread_safe=self.thread_safe,
        )
        return capture

    def get_imu_sample(self, timeout: int = TIMEOUT_WAIT_INFINITE) -> Optional["ImuSample"]:
        self._validate_is_opened()
        res, imu_sample = k4a_module.device_get_imu_sample(self._device_handle, self.thread_safe, timeout)
        _verify_error(res)
        return imu_sample

    @property
    def serial(self) -> str:
        self._validate_is_opened()
        ret = k4a_module.device_get_serialnum(self._device_handle, self.thread_safe)
        if ret == "":
            raise K4AException("Cannot read serial")
        return ret

    @property
    def calibration_raw(self) -> str:
        self._validate_is_opened()
        raw = k4a_module.device_get_raw_calibration(self._device_handle, self.thread_safe)
        return raw

    @calibration_raw.setter
    def calibration_raw(self, value: str):
        self._validate_is_opened()
        self._calibration = Calibration.from_raw(
            value, self._config.depth_mode, self._config.color_resolution, self.thread_safe
        )

    @property
    def sync_jack_status(self) -> Tuple[bool, bool]:
        self._validate_is_opened()
        res, jack_in, jack_out = k4a_module.device_get_sync_jack(self._device_handle, self.thread_safe)
        _verify_error(res)
        return jack_in == 1, jack_out == 1

    def _get_color_control(self, cmd: ColorControlCommand) -> Tuple[int, ColorControlMode]:
        self._validate_is_opened()
        res, mode, value = k4a_module.device_get_color_control(self._device_handle, self.thread_safe, cmd)
        _verify_error(res)
        return value, ColorControlMode(mode)

    def _set_color_control(self, cmd: ColorControlCommand, value: int, mode=ColorControlMode.MANUAL):
        self._validate_is_opened()
        res = k4a_module.device_set_color_control(self._device_handle, self.thread_safe, cmd, mode, value)
        _verify_error(res)

    @property
    def brightness(self) -> int:
        return self._get_color_control(ColorControlCommand.BRIGHTNESS)[0]

    @brightness.setter
    def brightness(self, value: int):
        self._set_color_control(ColorControlCommand.BRIGHTNESS, value)

    @property
    def contrast(self) -> int:
        return self._get_color_control(ColorControlCommand.CONTRAST)[0]

    @contrast.setter
    def contrast(self, value: int):
        self._set_color_control(ColorControlCommand.CONTRAST, value)

    @property
    def saturation(self) -> int:
        return self._get_color_control(ColorControlCommand.SATURATION)[0]

    @saturation.setter
    def saturation(self, value: int):
        self._set_color_control(ColorControlCommand.SATURATION, value)

    @property
    def sharpness(self) -> int:
        return self._get_color_control(ColorControlCommand.SHARPNESS)[0]

    @sharpness.setter
    def sharpness(self, value: int):
        self._set_color_control(ColorControlCommand.SHARPNESS, value)

    @property
    def backlight_compensation(self) -> int:
        return self._get_color_control(ColorControlCommand.BACKLIGHT_COMPENSATION)[0]

    @backlight_compensation.setter
    def backlight_compensation(self, value: int):
        self._set_color_control(ColorControlCommand.BACKLIGHT_COMPENSATION, value)

    @property
    def gain(self) -> int:
        return self._get_color_control(ColorControlCommand.GAIN)[0]

    @gain.setter
    def gain(self, value: int):
        self._set_color_control(ColorControlCommand.GAIN, value)

    @property
    def powerline_frequency(self) -> int:
        return self._get_color_control(ColorControlCommand.POWERLINE_FREQUENCY)[0]

    @powerline_frequency.setter
    def powerline_frequency(self, value: int):
        self._set_color_control(ColorControlCommand.POWERLINE_FREQUENCY, value)

    @property
    def exposure(self) -> int:
        # sets mode to manual
        return self._get_color_control(ColorControlCommand.EXPOSURE_TIME_ABSOLUTE)[0]

    @exposure.setter
    def exposure(self, value: int):
        self._set_color_control(ColorControlCommand.EXPOSURE_TIME_ABSOLUTE, value)

    @property
    def exposure_mode_auto(self) -> bool:
        return self._get_color_control(ColorControlCommand.EXPOSURE_TIME_ABSOLUTE)[1] == ColorControlMode.AUTO

    @exposure_mode_auto.setter
    def exposure_mode_auto(self, mode_auto: bool, value: int = 2500):
        mode = ColorControlMode.AUTO if mode_auto else ColorControlMode.MANUAL
        self._set_color_control(ColorControlCommand.EXPOSURE_TIME_ABSOLUTE, value=value, mode=mode)

    @property
    def whitebalance(self) -> int:
        # sets mode to manual
        return self._get_color_control(ColorControlCommand.WHITEBALANCE)[0]

    @whitebalance.setter
    def whitebalance(self, value: int):
        self._set_color_control(ColorControlCommand.WHITEBALANCE, value)

    @property
    def whitebalance_mode_auto(self) -> bool:
        return self._get_color_control(ColorControlCommand.WHITEBALANCE)[1] == ColorControlMode.AUTO

    @whitebalance_mode_auto.setter
    def whitebalance_mode_auto(self, mode_auto: bool, value: int = 2500):
        mode = ColorControlMode.AUTO if mode_auto else ColorControlMode.MANUAL
        self._set_color_control(ColorControlCommand.WHITEBALANCE, value=value, mode=mode)

    def _get_color_control_capabilities(self, cmd: ColorControlCommand) -> Optional["ColorControlCapabilities"]:
        self._validate_is_opened()
        res, capabilities = k4a_module.device_get_color_control_capabilities(self._device_handle, self.thread_safe, cmd)
        _verify_error(res)
        return capabilities

    def reset_color_control_to_default(self):
        for cmd in ColorControlCommand:
            capability = self._get_color_control_capabilities(cmd)
            self._set_color_control(cmd, capability["default_value"], capability["default_mode"])

    @property
    def calibration(self) -> Calibration:
        self._validate_is_opened()
        if not self._calibration:
            res, calibration_handle = k4a_module.device_get_calibration(
                self._device_handle, self.thread_safe, self._config.depth_mode, self._config.color_resolution
            )
            _verify_error(res)
            self._calibration = Calibration(
                handle=calibration_handle,
                depth_mode=self._config.depth_mode,
                color_resolution=self._config.color_resolution,
                thread_safe=self.thread_safe,
            )
        return self._calibration

    def _validate_is_opened(self):
        if not self.opened:
            raise K4AException("Device is not opened")


class ImuSample(TypedDict):
    temperature: float
    acc_sample: Tuple[float, float, float]
    acc_timestamp: int
    gyro_sample: Tuple[float, float, float]
    gyro_timestamp: int


class ColorControlCapabilities(TypedDict):
    color_control_command: ColorControlCommand
    supports_auto: bool
    min_value: int
    max_value: int
    step_value: int
    default_value: int
    default_mode: ColorControlMode



from typing import Optional

import numpy as np

import k4a_module

from .calibration import Calibration


def depth_image_to_color_camera(depth: np.ndarray, calibration: Calibration, thread_safe: bool) -> Optional[np.ndarray]:
    """
    Transform depth color_image to color color_image space
    Return empty result if transformation failed
    """
    return k4a_module.transformation_depth_image_to_color_camera(
        calibration.transformation_handle, thread_safe, depth, calibration.color_resolution,
    )


def depth_image_to_color_camera_custom(
    depth: np.ndarray, custom: np.ndarray, calibration: Calibration, thread_safe: bool, interp_nearest: bool = True,
) -> Optional[np.ndarray]:
    """
    Transforms depth image and custom image to color_image space
    Return empty result if transformation failed
    """
    return k4a_module.transformation_depth_image_to_color_camera_custom(
        calibration.transformation_handle, thread_safe, depth, custom, calibration.color_resolution, interp_nearest,
    )


def depth_image_to_point_cloud(
    depth: np.ndarray, calibration: Calibration, thread_safe: bool, calibration_type_depth=True
) -> Optional[np.ndarray]:
    """
    Transform depth color_image to point cloud
    Return empty result if transformation failed
    """
    return k4a_module.transformation_depth_image_to_point_cloud(
        calibration.transformation_handle, thread_safe, depth, calibration_type_depth,
    )


def color_image_to_depth_camera(
    color: np.ndarray, depth: np.ndarray, calibration: Calibration, thread_safe: bool
) -> Optional[np.ndarray]:
    """
    Transform color color_image to depth color_image space
    Return empty result if transformation failed
    """
    return k4a_module.transformation_color_image_to_depth_camera(
        calibration.transformation_handle, thread_safe, depth, color
    )



import sys
from enum import IntEnum
from pathlib import Path
from typing import Any, Optional, Tuple, Union


if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict

import k4a_module

from .calibration import Calibration
from .capture import PyK4ACapture
from .config import FPS, ColorResolution, DepthMode, ImageFormat, WiredSyncMode
from .errors import K4AException, _verify_error
from .results import Result, StreamResult


class SeekOrigin(IntEnum):
    BEGIN = 0
    END = 1
    DEVICE_TIME = 2


class Configuration(TypedDict):
    color_format: ImageFormat
    color_resolution: ColorResolution
    depth_mode: DepthMode
    camera_fps: FPS
    color_track_enabled: bool
    depth_track_enabled: bool
    ir_track_enabled: bool
    imu_track_enabled: bool
    depth_delay_off_color_usec: int
    wired_sync_mode: WiredSyncMode
    subordinate_delay_off_master_usec: int
    start_timestamp_offset_usec: int


class PyK4APlayback:
    def __init__(self, path: Union[str, Path], thread_safe: bool = True):
        self._path: Path = Path(path)
        self.thread_safe = thread_safe
        self._handle: Optional[object] = None
        self._length: Optional[int] = None
        self._calibration_json: Optional[str] = None
        self._calibration: Optional[Calibration] = None
        self._configuration: Optional[Configuration] = None

    def __del__(self):
        if self._handle:
            self.close()

    @property
    def path(self) -> Path:
        """
            Record file path
        """
        return self._path

    @property
    def configuration(self) -> Configuration:
        self._validate_is_open()
        if self._configuration is None:
            res, conf = k4a_module.playback_get_record_configuration(
                self._handle, self.thread_safe
            )  # type: int, Tuple[Any,...]
            _verify_error(res)
            self._configuration = Configuration(
                color_format=ImageFormat(conf[0]),
                color_resolution=ColorResolution(conf[1]),
                depth_mode=DepthMode(conf[2]),
                camera_fps=FPS(conf[3]),
                color_track_enabled=bool(conf[4]),
                depth_track_enabled=bool(conf[5]),
                ir_track_enabled=bool(conf[6]),
                imu_track_enabled=bool(conf[7]),
                depth_delay_off_color_usec=conf[8],
                wired_sync_mode=WiredSyncMode(conf[9]),
                subordinate_delay_off_master_usec=conf[10],
                start_timestamp_offset_usec=conf[11],
            )
        return self._configuration

    @property
    def length(self) -> int:
        """
            Record length in usec
        """
        if self._length is None:
            self._validate_is_open()
            self._length = k4a_module.playback_get_recording_length_usec(self._handle, self.thread_safe)
        return self._length

    @property
    def calibration_raw(self) -> str:
        self._validate_is_open()
        res, raw = k4a_module.playback_get_raw_calibration(self._handle, self.thread_safe)
        _verify_error(res)
        return raw

    @calibration_raw.setter
    def calibration_raw(self, value: str):
        self._validate_is_open()
        self._calibration = Calibration.from_raw(
            value, self.calibration.depth_mode, self.calibration.color_resolution, self.thread_safe
        )

    @property
    def calibration(self) -> Calibration:
        self._validate_is_open()
        if self._calibration is None:
            res, handle = k4a_module.playback_get_calibration(self._handle, self.thread_safe)
            _verify_error(res)
            self._calibration = Calibration(
                handle=handle,
                depth_mode=self.configuration["depth_mode"],
                color_resolution=self.configuration["color_resolution"],
                thread_safe=self.thread_safe,
            )
        return self._calibration

    def open(self) -> None:
        """
            Open record file
        """
        if self._handle:
            raise K4AException("Playback already opened")
        result, handle = k4a_module.playback_open(str(self._path), self.thread_safe)
        if Result(result) != Result.Success:
            raise K4AException(f"Cannot open file {self._path}")

        self._handle = handle

    def close(self):
        """
            Close record file
        """
        self._validate_is_open()
        k4a_module.playback_close(self._handle, self.thread_safe)
        self._handle = None

    def seek(self, offset: int, origin: SeekOrigin = SeekOrigin.BEGIN) -> None:
        """
            Seek playback pointer to specified offset
        """
        self._validate_is_open()
        result = k4a_module.playback_seek_timestamp(self._handle, self.thread_safe, offset, int(origin))
        self._verify_stream_error(result)

    def get_next_capture(self):
        self._validate_is_open()
        result, capture_handle = k4a_module.playback_get_next_capture(self._handle, self.thread_safe)
        self._verify_stream_error(result)
        return PyK4ACapture(
            calibration=self.calibration,
            capture_handle=capture_handle,
            color_format=self.configuration["color_format"],
            thread_safe=self.thread_safe,
        )

    def get_previouse_capture(self):
        self._validate_is_open()
        result, capture_handle = k4a_module.playback_get_previous_capture(self._handle, self.thread_safe)
        self._verify_stream_error(result)
        return PyK4ACapture(
            calibration=self.calibration,
            capture_handle=capture_handle,
            color_format=self.configuration["color_format"],
            thread_safe=self.thread_safe,
        )

    def _validate_is_open(self):
        if not self._handle:
            raise K4AException("Playback not opened.")

    @staticmethod
    def _verify_stream_error(res: int):
        """
        Validate k4a_module result(k4a_stream_result_t)
        """
        result = StreamResult(res)
        if result == StreamResult.Failed:
            raise K4AException()
        elif result == StreamResult.EOF:
            raise EOFError()



from typing import Optional

import numpy as np

import k4a_module

from .calibration import Calibration
from .config import ImageFormat
from .errors import K4AException
from .transformation import (
    color_image_to_depth_camera,
    depth_image_to_color_camera,
    depth_image_to_color_camera_custom,
    depth_image_to_point_cloud,
)


class PyK4ACapture:
    def __init__(
        self, calibration: Calibration, capture_handle: object, color_format: ImageFormat, thread_safe: bool = True
    ):
        self._calibration: Calibration = calibration
        self._capture_handle: object = capture_handle  # built-in PyCapsule
        self.thread_safe = thread_safe
        self._color_format = color_format

        self._color: Optional[np.ndarray] = None
        self._color_timestamp_usec: int = 0
        self._color_system_timestamp_nsec: int = 0
        self._color_exposure_usec: Optional[int] = None
        self._color_white_balance: Optional[int] = None
        self._depth: Optional[np.ndarray] = None
        self._depth_timestamp_usec: int = 0
        self._depth_system_timestamp_nsec: int = 0
        self._ir: Optional[np.ndarray] = None
        self._ir_timestamp_usec: int = 0
        self._ir_system_timestamp_nsec: int = 0
        self._depth_point_cloud: Optional[np.ndarray] = None
        self._transformed_depth: Optional[np.ndarray] = None
        self._transformed_depth_point_cloud: Optional[np.ndarray] = None
        self._transformed_color: Optional[np.ndarray] = None
        self._transformed_ir: Optional[np.ndarray] = None

    @property
    def color(self) -> Optional[np.ndarray]:
        if self._color is None:
            (
                self._color,
                self._color_timestamp_usec,
                self._color_system_timestamp_nsec,
            ) = k4a_module.capture_get_color_image(self._capture_handle, self.thread_safe)
        return self._color

    @property
    def color_timestamp_usec(self) -> int:
        """Device timestamp for color image. Not equal host machine timestamp!"""
        if self._color is None:
            self.color
        return self._color_timestamp_usec

    @property
    def color_system_timestamp_nsec(self) -> int:
        """System timestamp for color image in nanoseconds. Corresponds to Python's time.perf_counter_ns()."""
        if self._color is None:
            self.color
        return self._color_system_timestamp_nsec

    @property
    def color_exposure_usec(self) -> int:
        if self._color_exposure_usec is None:
            value = k4a_module.color_image_get_exposure_usec(self._capture_handle)
            if value == 0:
                raise K4AException("Cannot read exposure from color image")
            self._color_exposure_usec = value
        return self._color_exposure_usec

    @property
    def color_white_balance(self) -> int:
        if self._color_white_balance is None:
            value = k4a_module.color_image_get_white_balance(self._capture_handle)
            if value == 0:
                raise K4AException("Cannot read white balance from color image")
            self._color_white_balance = value
        return self._color_white_balance

    @property
    def depth(self) -> Optional[np.ndarray]:
        if self._depth is None:
            (
                self._depth,
                self._depth_timestamp_usec,
                self._depth_system_timestamp_nsec,
            ) = k4a_module.capture_get_depth_image(self._capture_handle, self.thread_safe)
        return self._depth

    @property
    def depth_timestamp_usec(self) -> int:
        """Device timestamp for depth image. Not equal host machine timestamp!. Like as equal IR image timestamp"""
        if self._depth is None:
            self.depth
        return self._depth_timestamp_usec

    @property
    def depth_system_timestamp_nsec(self) -> int:
        """System timestamp for depth image in nanoseconds. Corresponds to Python's time.perf_counter_ns()."""
        if self._depth is None:
            self.depth
        return self._depth_system_timestamp_nsec

    @property
    def ir(self) -> Optional[np.ndarray]:
        """Device timestamp for IR image. Not equal host machine timestamp!. Like as equal depth image timestamp"""
        if self._ir is None:
            self._ir, self._ir_timestamp_usec, self._ir_system_timestamp_nsec = k4a_module.capture_get_ir_image(
                self._capture_handle, self.thread_safe
            )
        return self._ir

    @property
    def ir_timestamp_usec(self) -> int:
        if self._ir is None:
            self.ir
        return self._ir_timestamp_usec

    @property
    def ir_system_timestamp_nsec(self) -> int:
        """System timestamp for IR image in nanoseconds. Corresponds to Python's time.perf_counter_ns()."""
        if self._ir is None:
            self.ir
        return self._ir_system_timestamp_nsec

    @property
    def transformed_depth(self) -> Optional[np.ndarray]:
        if self._transformed_depth is None and self.depth is not None:
            self._transformed_depth = depth_image_to_color_camera(self._depth, self._calibration, self.thread_safe)
        return self._transformed_depth

    @property
    def depth_point_cloud(self) -> Optional[np.ndarray]:
        if self._depth_point_cloud is None and self.depth is not None:
            self._depth_point_cloud = depth_image_to_point_cloud(
                self._depth, self._calibration, self.thread_safe, calibration_type_depth=True,
            )
        return self._depth_point_cloud

    @property
    def transformed_depth_point_cloud(self) -> Optional[np.ndarray]:
        if self._transformed_depth_point_cloud is None and self.transformed_depth is not None:
            self._transformed_depth_point_cloud = depth_image_to_point_cloud(
                self.transformed_depth, self._calibration, self.thread_safe, calibration_type_depth=False,
            )
        return self._transformed_depth_point_cloud

    @property
    def transformed_color(self) -> Optional[np.ndarray]:
        if self._transformed_color is None and self.depth is not None and self.color is not None:
            if self._color_format != ImageFormat.COLOR_BGRA32:
                raise RuntimeError(
                    "color color_image must be of color_format K4A_IMAGE_FORMAT_COLOR_BGRA32 for "
                    "transformation_color_image_to_depth_camera"
                )
            self._transformed_color = color_image_to_depth_camera(
                self.color, self.depth, self._calibration, self.thread_safe
            )
        return self._transformed_color

    @property
    def transformed_ir(self) -> Optional[np.ndarray]:
        if self._transformed_ir is None and self.ir is not None and self.depth is not None:
            result = depth_image_to_color_camera_custom(self.depth, self.ir, self._calibration, self.thread_safe)
            if result is None:
                return None
            else:
                self._transformed_ir, self._transformed_depth = result
        return self._transformed_ir



from enum import IntEnum
from typing import Tuple


# k4a_fps_t
class FPS(IntEnum):
    FPS_5 = 0
    FPS_15 = 1
    FPS_30 = 2


# k4a_image_format_t
class ImageFormat(IntEnum):
    COLOR_MJPG = 0
    COLOR_NV12 = 1
    COLOR_YUY2 = 2
    COLOR_BGRA32 = 3
    DEPTH16 = 4
    IR16 = 5
    CUSTOM8 = 6
    CUSTOM16 = 7
    CUSTOM = 8


# k4a_depth_mode_t
class DepthMode(IntEnum):
    OFF = 0
    NFOV_2X2BINNED = 1
    NFOV_UNBINNED = 2
    WFOV_2X2BINNED = 3
    WFOV_UNBINNED = 4
    PASSIVE_IR = 5


# k4a_color_resolution_t
class ColorResolution(IntEnum):
    OFF = 0
    RES_720P = 1
    RES_1080P = 2
    RES_1440P = 3
    RES_1536P = 4
    RES_2160P = 5
    RES_3072P = 6


# k4a_wired_sync_mode_t
class WiredSyncMode(IntEnum):
    STANDALONE = 0
    MASTER = 1
    SUBORDINATE = 2


# k4a_color_control_command_t
class ColorControlCommand(IntEnum):
    EXPOSURE_TIME_ABSOLUTE = 0
    AUTO_EXPOSURE_PRIORITY = 1  # deprecated
    BRIGHTNESS = 2
    CONTRAST = 3
    SATURATION = 4
    SHARPNESS = 5
    WHITEBALANCE = 6
    BACKLIGHT_COMPENSATION = 7
    GAIN = 8
    POWERLINE_FREQUENCY = 9


# k4a_color_control_mode_t
class ColorControlMode(IntEnum):
    AUTO = 0
    MANUAL = 1


class Config:
    def __init__(
        self,
        color_resolution: ColorResolution = ColorResolution.RES_720P,
        color_format: ImageFormat = ImageFormat.COLOR_BGRA32,
        depth_mode: DepthMode = DepthMode.NFOV_UNBINNED,
        camera_fps: FPS = FPS.FPS_30,
        synchronized_images_only: bool = True,
        depth_delay_off_color_usec: int = 0,
        wired_sync_mode: WiredSyncMode = WiredSyncMode.STANDALONE,
        subordinate_delay_off_master_usec: int = 0,
        disable_streaming_indicator: bool = False,
    ):
        self.color_resolution = color_resolution
        self.color_format = color_format
        self.depth_mode = depth_mode
        self.camera_fps = camera_fps
        self.synchronized_images_only = synchronized_images_only
        self.depth_delay_off_color_usec = depth_delay_off_color_usec
        self.wired_sync_mode = wired_sync_mode
        self.subordinate_delay_off_master_usec = subordinate_delay_off_master_usec
        self.disable_streaming_indicator = disable_streaming_indicator
        assert self.subordinate_delay_off_master_usec >= 0

    def unpack(self) -> Tuple[ImageFormat, ColorResolution, DepthMode, FPS, bool, int, WiredSyncMode, int, bool]:
        return (
            self.color_format,
            self.color_resolution,
            self.depth_mode,
            self.camera_fps,
            self.synchronized_images_only,
            self.depth_delay_off_color_usec,
            self.wired_sync_mode,
            self.subordinate_delay_off_master_usec,
            self.disable_streaming_indicator,
        )



from setuptools import setup, Extension

import sys
if sys.version_info[0] == 2:
    sys.exit("Python 2 is not supported.")

# Enables --editable install with --user
# https://github.com/pypa/pip/issues/7953
import site
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

# Bypass import numpy before running install_requires
# https://stackoverflow.com/questions/54117786/add-numpy-get-include-argument-to-setuptools-without-preinstalled-numpy
class get_numpy_include:
    def __str__(self):
        import numpy
        return numpy.get_include()


module = Extension('k4a_module',
                   sources=['pyk4a/pyk4a.cpp'],
                   include_dirs=[get_numpy_include()],
                   libraries=['k4a', 'k4arecord'])

setup(ext_modules=[module])
